@@ -1,7 +1,7 @@
-/* @(#)sdd.c	1.31 01/07/17 Copyright 1984-2001 J. Schilling */
+/* @(#)sdd.c	1.52 04/07/27 Copyright 1984-2004 J. Schilling */
 #ifndef lint
 static	char sccsid[] =
-	"@(#)sdd.c	1.31 01/07/17 Copyright 1984-2001 J. Schilling";
+	"@(#)sdd.c	1.52 04/07/27 Copyright 1984-2004 J. Schilling";
 #endif
 /*
  *	sdd	Disk and Tape copy
@@ -15,7 +15,7 @@ static	char sccsid[] =
  *	than large files, we use long long for several important
  *	variables that should not overflow.
  *
- *	Copyright (c) 1984-2001 J. Schilling
+ *	Copyright (c) 1984-2004 J. Schilling
  */
 /*
  * This program is free software; you can redistribute it and/or modify
@@ -28,12 +28,27 @@ static	char sccsid[] =
  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  * GNU General Public License for more details.
  *
- * You should have received a copy of the GNU General Public License
- * along with this program; see the file COPYING.  If not, write to
- * the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
+ * You should have received a copy of the GNU General Public License along with
+ * this program; see the file COPYING.  If not, write to the Free Software
+ * Foundation, 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
  */
 
 #include <mconfig.h>
+
+/*
+ * XXX Until we find a better way, the next definitions must be in sync
+ * XXX with the definitions in librmt/remote.c
+ */
+#if !defined(HAVE_FORK) || !defined(HAVE_SOCKETPAIR) || !defined(HAVE_DUP2)
+#undef	USE_RCMD_RSH
+#endif
+#if !defined(HAVE_GETSERVBYNAME)
+#undef	USE_REMOTE				/* Cannot get rcmd() port # */
+#endif
+#if (!defined(HAVE_NETDB_H) || !defined(HAVE_RCMD)) && !defined(USE_RCMD_RSH)
+#undef	USE_REMOTE				/* There is no rcmd() */
+#endif
+
 #include <stdio.h>
 #include <stdxlib.h>
 #include <unixstd.h>
@@ -44,7 +59,7 @@ static	char sccsid[] =
 #include <fctldefs.h>
 #include <strdefs.h>
 #include <schily.h>
-#include "remote.h"
+#include <librmt.h>
 
 #ifdef	SIGRELSE
 #	define	signal	sigset	/* reliable signal */
@@ -57,37 +72,38 @@ static	char sccsid[] =
 #	define	ENDOFFILE	EFBIG
 #endif
 
-EXPORT	int	main		__PR((int ac, char** av));
+EXPORT	int	main		__PR((int ac, char **av));
+LOCAL	void	set_signal	__PR((int sig, RETSIGTYPE (*handler)(int)));
 LOCAL	void	intr		__PR((int sig));
-LOCAL	FILE*	openfile	__PR((char* name, char* mode));
-LOCAL	char*	memalloc	__PR((long size));
+LOCAL	FILE 	*openfile	__PR((char *name, char *mode));
+LOCAL	char	*memalloc	__PR((long size));
 LOCAL	void	simple_copy	__PR((void));
 LOCAL	void	copy_reblocked	__PR((void));
-LOCAL	long	readvol		__PR((char* buf, long len));
-LOCAL	long	writevol	__PR((char* buf, long len));
+LOCAL	long	readvol		__PR((char *buf, long len));
+LOCAL	long	writevol	__PR((char *buf, long len));
 LOCAL	BOOL	next_in		__PR((void));
 LOCAL	BOOL	next_out	__PR((void));
-LOCAL	BOOL	next		__PR((char* fname, char* inout, int fd, long pos, int volnum));
-LOCAL	BOOL	cont		__PR((char* inout, int num));
-LOCAL	void	makelower	__PR((char* s));
-LOCAL	long	readbuf		__PR((char* buf, long len));
-LOCAL	long	writebuf	__PR((char* buf, long len));
-LOCAL	long	readblocks	__PR((char* buf, long len));
-LOCAL	long	writeblocks	__PR((char* buf, long len));
-LOCAL	void	fill		__PR((char* bp, long start, long end));
-LOCAL	void	swabb		__PR((char* bp, int cnt));
-LOCAL	void	lcase		__PR((char* bp, long cnt));
-LOCAL	void	ucase		__PR((char* bp, long cnt));
-LOCAL	long	block		__PR((char* bp, long cnt));
-LOCAL	long	unblock		__PR((char* bp, long cnt));
-LOCAL	void	conv		__PR((char* bp, long cnt, unsigned char* tab));
+LOCAL	BOOL	next		__PR((char *fname, char *inout, int fd, long pos, int volnum));
+LOCAL	BOOL	cont		__PR((char *inout, int num));
+LOCAL	void	makelower	__PR((char *s));
+LOCAL	long	readbuf		__PR((char *buf, long len));
+LOCAL	long	writebuf	__PR((char *buf, long len));
+LOCAL	long	readblocks	__PR((char *buf, long len));
+LOCAL	long	writeblocks	__PR((char *buf, long len));
+LOCAL	void	fill		__PR((char *bp, long start, long end));
+LOCAL	void	swabb		__PR((char *bp, int cnt));
+LOCAL	void	lcase		__PR((char *bp, long cnt));
+LOCAL	void	ucase		__PR((char *bp, long cnt));
+LOCAL	long	block		__PR((char *bp, long cnt));
+LOCAL	long	unblock		__PR((char *bp, long cnt));
+LOCAL	void	conv		__PR((char *bp, long cnt, unsigned char *tab));
 LOCAL	void	term		__PR((int ret));
 LOCAL	void	getstarttime	__PR((void));
 LOCAL	void	prstats		__PR((void));
-LOCAL	void	getopts		__PR((int ac, char** av));
-LOCAL	Llong	number		__PR((char* arg, int* retp));
-LOCAL	int	getnum		__PR((char* arg, long* valp));
-LOCAL	int	getllnum	__PR((char* arg, Llong* valp));
+LOCAL	void	getopts		__PR((int ac, char **av));
+LOCAL	Llong	number		__PR((char *arg, int *retp));
+LOCAL	int	getnum		__PR((char *arg, long *valp));
+LOCAL	int	getllnum	__PR((char *arg, Llong *valp));
 LOCAL	void	usage		__PR((int ex));
 LOCAL	int	openremote	__PR((char *filename, int iosize));
 LOCAL	int	ropenfile	__PR((int rfd, char *name, int mode));
@@ -97,9 +113,9 @@ LOCAL	off_t	riseek		__PR((off_t pos));
 LOCAL	off_t	roseek		__PR((off_t pos));
 LOCAL	off_t	ifsize		__PR((void));
 
-#define min(a,b)	(a < b ? a : b)
+#define	min(a, b)	(a < b ? a : b)
 
-#define	BSIZE		512L
+#define	SDD_BSIZE	512L
 
 #define	FILL		000001
 #define	SWAB		000002
@@ -420,10 +436,19 @@ main(ac, av)
 	tty = stdin;
 
 	getstarttime();
-	if (signal(SIGINT, intr) == SIG_IGN)
-		(void)signal(SIGINT, SIG_IGN);
-	if (signal(SIGQUIT, intr) == SIG_IGN)
-		(void)signal(SIGQUIT, SIG_IGN);
+	if (signal(SIGINT, SIG_IGN) != SIG_IGN)
+		(void) set_signal(SIGINT, intr);
+	if (signal(SIGQUIT, SIG_IGN) != SIG_IGN)
+		(void) set_signal(SIGQUIT, intr);
+#ifdef	SIGINFO
+	/*
+	 * Be polite to *BSD users.
+	 * They copied our idea and implemented intermediate status
+	 * printing in 'dd' in 1990.
+	 */
+	if (signal(SIGINFO, SIG_IGN) != SIG_IGN)
+		(void) set_signal(SIGINFO, intr);
+#endif
 
 #ifdef	USE_REMOTE
 	rmtdebug(debug);
@@ -432,10 +457,10 @@ main(ac, av)
 	if (outfile)
 		rmtout = rmtfilename(outfile);
 	if (rmtin)
-		rmtifd = openremote(infile, ibs); /* This needs super user privilleges */
+		rmtifd = openremote(infile, ibs); /* Needs root privilleges */
 
 	if (rmtout)
-		rmtofd = openremote(outfile, obs); /* This needs super user privilleges */
+		rmtofd = openremote(outfile, obs); /* Needs root privilleges */
 #endif
 
 
@@ -450,12 +475,12 @@ main(ac, av)
 		if (setuid(getuid()) < 0)
 #endif
 #endif
-			comerr("Panic cannot set back efective uid.\n");
+			comerr("Panic cannot set back effective uid.\n");
 	}
 
 	if (infile) {
 		if (rmtin) {
-	 		ropenfile(rmtifd, rmtin, O_RDONLY);
+			ropenfile(rmtifd, rmtin, O_RDONLY);
 		} else {
 			fin = openfile(infile, "ru");
 		}
@@ -469,9 +494,13 @@ main(ac, av)
 	}
 	if (outfile) {
 		if (rmtout) {
-	 		ropenfile(rmtofd, rmtout, notrunc ? (O_WRONLY|O_CREAT):(O_WRONLY|O_CREAT|O_TRUNC));
+			ropenfile(rmtofd, rmtout,
+					notrunc ?
+					(O_WRONLY|O_CREAT):
+					(O_WRONLY|O_CREAT|O_TRUNC));
 		} else {
-			fout = openfile(outfile, notrunc ? "wcu" : "wctu");
+			fout = openfile(outfile,
+					notrunc ? "wcu" : "wctu");
 		}
 	} else {
 		fout = stdout;
@@ -495,42 +524,51 @@ main(ac, av)
 
 	getstarttime();
 
-	if (obs != ibs
-	    || (flags & (BLOCK|UNBLOCK))	/* Reblock forced uncond.    */
-	    || (ivsize && !(flags & NULLOUT))	/* Reblock at end of vol     */
-	    || (ovsize &&
-	        ((ovsize - ovpos) % obs		/* Reblock at end of 1st vol */
-		  || (ovsize - ovseek) % obs)))	/* Reblock at end of 2nd vol */
+	if ((obs != ibs) ||
+	    (flags & (BLOCK|UNBLOCK)) ||	/* Reblock forced uncond.   */
+	    (ivsize && !(flags & NULLOUT)) ||	/* Reblock at end of vol    */
+	    (ovsize &&
+	    (((ovsize - ovpos) % obs) ||	/* Reblock at end of 1st vol */
+	    ((ovsize - ovseek) % obs))))	/* Reblock at end of 2nd vol */
 		copy_reblocked();
 	else
 		simple_copy();
 
-	if (rmtout) {
-#ifdef	USE_REMOTE
-		/*
-		 * Cannot happen in non remote versions.
-		 */
-		if (rmtclose(rmtofd) < 0)
-			ret = geterrno();
-#endif
-	} else {
-#ifdef	HAVE_FSYNC
-		if (fsync(ofd) != 0)
-			ret = geterrno();
-#endif
-		if (close(ofd) != 0 && ret == 0)
-			ret = geterrno();
-	}
-
 	term(ret);
 	return (0);	/* Keep lint happy */
 }
 
+LOCAL void
+set_signal(sig, handler)
+	int		sig;
+	RETSIGTYPE	(*handler)	__PR((int));
+{
+#if	defined(HAVE_SIGPROCMASK) && defined(SA_RESTART)
+	struct sigaction sa;
+
+	sigemptyset(&sa.sa_mask);
+	sa.sa_handler = handler;
+	sa.sa_flags = SA_RESTART;
+	(void) sigaction(sig, &sa, (struct sigaction *)0);
+#else
+#ifdef	HAVE_SIGSETMASK
+	struct sigvec	sv;
+
+	sv.sv_mask = 0;
+	sv.sv_handler = handler;
+	sv.sv_flags = 0;
+	(void) sigvec(sig, &sv, (struct sigvec *)0);
+#else
+	(void) signal(sig, handler);
+#endif
+#endif
+}
+
 LOCAL void
 intr(sig)
 	int	sig;
 {
-	(void)signal(sig, intr);
+	(void) signal(sig, intr);
 	prstats();
 	if (sig == SIGINT) {
 		errmsgno(EX_BAD, "KILLED by SIGINT.\n");
@@ -545,7 +583,7 @@ openfile(name, mode)
 {
 	FILE	*f;
 
-	if ((f = fileopen(name, mode)) == (FILE*) NULL)
+	if ((f = fileopen(name, mode)) == (FILE *) NULL)
 		comerr("Can't open '%s'.\n", name);
 	file_raise(f, FALSE);
 	return (f);
@@ -600,19 +638,19 @@ simple_copy()
 	if (rflags & NULLIN)
 		fill(obp, 0L, bs);
 
-	while((cnt = readvol(obp, bs)) > 0) {
+	while ((cnt = readvol(obp, bs)) > 0) {
 		if (rflags & NULLOUT) {
 			if (progress && !debug) {
-				(void)putc('.', stderr);
-				(void)fflush(stderr);
+				(void) putc('.', stderr);
+				(void) fflush(stderr);
 			}
 			continue;
 		}
-		if(cnt < bs && (rflags & FILL)) {
+		if (cnt < bs && (rflags & FILL)) {
 			fill(obp, cnt, bs);
 			cnt = bs;
 		}
-		for (obcnt = 0; obcnt < cnt ;)
+		for (obcnt = 0; obcnt < cnt; )
 			obcnt += writevol(obp + obcnt, (long) (cnt - obcnt));
 	}
 }
@@ -631,14 +669,14 @@ copy_reblocked()
 	register char	*obp;
 
 	if (debug)
-		error ("Copy reblocked ...\n");
+		error("Copy reblocked ...\n");
 
 	obp = memalloc((long) (obs + ibs));
 
 	while ((cnt = readvol(obp + obcnt, ibs)) > 0) {
 		obcnt += cnt;
 		cnt = 0;
-		for (cnt = 0; obcnt - cnt >= obs ;)
+		for (cnt = 0; obcnt - cnt >= obs; )
 			cnt += writevol(obp + cnt, obs);
 		if (cnt != 0) {
 			if (obcnt > cnt) {
@@ -653,7 +691,7 @@ copy_reblocked()
 	if (obcnt > 0) {
 		if (debug)
 			error("Writing tail of %ld bytes\n", obcnt);
-		if(flags & FILL) {
+		if (flags & FILL) {
 			fill(obp, obcnt, obs);
 			obcnt = obs;
 		}
@@ -687,9 +725,9 @@ readvol(buf, len)
 		return (len);
 	}
 
-	if (ivsize == 0)
+	if (ivsize == 0) {
 		cnt = readbuf(buf, len);
-	else for(;;) {
+	} else for (;;) {
 		if ((left = ivsize - ivpos) > 0) {
 			if ((cnt = readbuf(buf, (long) min(len, left))) > 0)
 				break;
@@ -779,17 +817,17 @@ next(fname, inout, fd, pos, volnum)
 	int	volnum;
 {
 	if (progress || debug) {
-		(void)putc('\n', stderr);
-		(void)fflush(stderr);
+		(void) putc('\n', stderr);
+		(void) fflush(stderr);
 	}
-	errmsgno (EX_BAD, "Done with %s volume # %d.\n", inout, volnum);
+	errmsgno(EX_BAD, "Done with %s volume # %d.\n", inout, volnum);
 	if (!cont(inout, ++volnum))
 		return (FALSE);
 	error("Insert %s volume # %d in '%s' and then hit <cr>: ",
 						inout, volnum, fname);
-	(void)fflush(stderr);
+	(void) fflush(stderr);
 	while (getc(tty) != '\n')
-		 if (feof(tty))
+		if (feof(tty))
 			return (FALSE);
 	error("Working on %s volume # %d of '%s'.\n", inout, volnum, fname);
 	if (fd == ifd)
@@ -807,10 +845,10 @@ cont(inout, num)
 		char	answer [16];
 	register char	*ap;
 
-	for(;;) {
+	for (;;) {
 		error("Do you want to continue with %s volume # %d (y/n): ",
 						inout, num);
-		(void)fflush(stderr);
+		(void) fflush(stderr);
 		ap = answer;
 		if (fgetline(tty, ap, 16) == EOF)
 			return (FALSE);
@@ -843,7 +881,7 @@ makelower(s)
  * Read one input block.
  * Call readblocks if an error occured and -noerror has been specified.
  */
-LOCAL long 
+LOCAL long
 readbuf(buf, len)
 	char	*buf;
 	long	len;
@@ -854,7 +892,7 @@ readbuf(buf, len)
 	lastreaderrs = (Llong)0;
 	if (debug) {
 		error("readbuf  (%d, %p, %ld) ", ifd, (void *)buf, len);
-		(void)fflush(stderr);
+		(void) fflush(stderr);
 	}
 	if (noerror && noseek)
 		fill(buf, 0L, len);
@@ -864,21 +902,25 @@ readbuf(buf, len)
 			err = geterrno();
 		error("= %ld\n", cnt);
 	}
-	if(cnt < 0) {
+	if (cnt < 0) {
 		if (!debug)
 			err = geterrno();
 		if (progress && !debug)
-			(void)putc('\n', stderr);
+			(void) putc('\n', stderr);
 		errmsgno(err, "Error reading '%s'.\n", infile);
-		if(noerror)
+#ifdef	ECONNRESET
+		if (noerror && err != EPIPE && err != ECONNRESET)
+#else
+		if (noerror && err != EPIPE)
+#endif
 			cnt = readblocks(buf, len);
 		else
 			term(err);
-  	}
+	}
 	if (cnt == 0)
-		if(count != 0 && ivsize == 0) {
+		if (count != 0 && ivsize == 0) {
 			if (progress || debug)
-				(void)putc('\n', stderr);
+				(void) putc('\n', stderr);
 			errmsgno(EX_BAD, "END OF FILE\n");
 		}
 	return (cnt);
@@ -886,14 +928,15 @@ readbuf(buf, len)
 
 /*
  * Write one output block.
- * Call writwblocks if an error occured and -noerror has been specified.
+ * Call writeblocks if an error occured and -noerror has been specified.
  */
 LOCAL long
 writebuf(buf, len)
 	char	*buf;
 	long	len;
 {
-	long cnt;
+	long	cnt;
+	int	err = 0;
 
 	if (debug)
 		error("writebuf (%d, %p, %ld)\n", ofd, (void *)buf, len);
@@ -901,28 +944,33 @@ writebuf(buf, len)
 		if (debug)
 			error("seek(%d, %lld)\n", ofd, (Llong)(ovpos + len));
 		if (roseek(ovpos + len) == (off_t)-1) {
-			cnt = geterrno();
+			err = geterrno();
 			if (progress && !debug)
-				(void)putc('\n', stderr);
-			errmsgno((int) cnt, "Error seeking '%s'.\n", outfile);
+				(void) putc('\n', stderr);
+			errmsgno((int) err, "Error seeking '%s'.\n", outfile);
 		}
 		cnt = len;
-	} else if((cnt = rwrite(buf, (int)len)) <= 0) {
+	} else if ((cnt = rwrite(buf, (int)len)) <= 0) {
 		if (cnt == 0)		/* EOF */
-			cnt = ENDOFFILE;
+			err = ENDOFFILE;
 		else if (cnt < 0)
-			cnt = geterrno();
+			err = geterrno();
 		if (progress && !debug)
-			(void)putc('\n', stderr);
-		errmsgno((int) cnt, "Error writing '%s'.\n", outfile);
-		if (noerror && cnt != ENDOFFILE)
+			(void) putc('\n', stderr);
+		errmsgno((int) err, "Error writing '%s'.\n", outfile);
+		if (noerror &&
+#ifdef	ECONNRESET
+		    err != ENDOFFILE && err != EPIPE && err != ECONNRESET)
+#else
+		    err != ENDOFFILE && err != EPIPE)
+#endif
 			cnt = writeblocks(buf, len);
 		else
-			term((int) cnt);
+			term((int) err);
 	}
 	if (progress && !debug) {
-		(void)putc('.', stderr);
-		(void)fflush(stderr);
+		(void) putc('.', stderr);
+		(void) fflush(stderr);
 	}
 	return (cnt);
 }
@@ -940,53 +988,67 @@ readblocks(buf, len)
 	register int	trys;
 	register off_t	pos = ivpos;
 	register long	total = 0;
+		int	err = 0;
 
 	if (noseek) {
-		errmsgno(EX_BAD, "Can't read %ld Bytes at %lld\n", len, (Llong)pos);
+		errmsgno(EX_BAD, "Can't read %ld Bytes at %lld\n",
+							len, (Llong)pos);
 		readerrs++;
 		return (len);
 	}
-	while(len > 0) {
-		aktlen = min(BSIZE, len);
+	while (len > 0) {
+		aktlen = min(SDD_BSIZE, len);
 		trys = 0;
 		do {
-			if (trys && !(trys & 15)){
+			if (trys && !(trys & 15)) {
 				if (debug) {
-					(void)putc('+', stderr);
-					(void)fflush(stderr);
+					(void) putc('+', stderr);
+					(void) fflush(stderr);
 				}
-				(void) riseek((off_t)(ifsize()- BSIZE));
+				(void) riseek((off_t)(ifsize() - SDD_BSIZE));
 				(void) rread(buf, (int) aktlen);
-			}
-			else if (trys > 0 && (trys == 2 || ! (trys & 7))) {
+
+			} else if (trys > 0 && (trys == 2 || ! (trys & 7))) {
 				if (debug) {
-					(void)putc('-', stderr);
-					(void)fflush(stderr);
+					(void) putc('-', stderr);
+					(void) fflush(stderr);
 				}
 				(void) riseek((off_t)0);
 				(void) rread(buf, (int) aktlen);
 			}
 			if (debug) {
-				(void)putc(',', stderr);
-				(void)fflush(stderr);
+				(void) putc(',', stderr);
+				(void) fflush(stderr);
 			}
 			fill(buf, 0L, aktlen);
 			(void) riseek(pos);
 			cnt = rread(buf, (int) aktlen);
+			if (cnt < 0) {
+				err = geterrno();
+#ifdef	ECONNRESET
+				if (err == EPIPE || err == ECONNRESET)
+#else
+				if (err == EPIPE)
+#endif
+					break;
+				err = 0;
+			}
 		} while (cnt < 0 && trys++ < try);
 
-		if(cnt < 0) {
+		if (cnt < 0) {
 			if (progress || debug) {
-				(void)putc('\n', stderr);
-				(void)fflush(stderr);
+				(void) putc('\n', stderr);
+				(void) fflush(stderr);
 			}
 			errmsgno(EX_BAD, "Block %lld not read correctly.\n",
-								(Llong)pos/BSIZE);
+						(Llong)pos/(Llong)SDD_BSIZE);
+			if (err != 0)
+				term((int) err);
 			cnt = aktlen;
 			readerrs++;
 			lastreaderrs++;
-		}
-		else if (cnt == 0)	/* EOF */
+
+		} else if (cnt == 0)	/* EOF */
 			break;
 		buf += cnt;
 		len -= cnt;
@@ -1010,51 +1072,67 @@ writeblocks(buf, len)
 	register int	trys;
 	register off_t	pos = ovpos;
 	register long	total = 0;
-		char	rdbuf[BSIZE];
+/*		char	rdbuf[SDD_BSIZE];*/
+		int	err = 0;
 
 	if (noseek) {
-		errmsgno(EX_BAD, "Can't write %ld Bytes at %lld\n", len, (Llong)pos);
+		errmsgno(EX_BAD, "Can't write %ld Bytes at %lld\n",
+							len, (Llong)pos);
 		writeerrs++;
 		return (len);
 	}
-	while(len > 0) {
-		aktlen = min(BSIZE, len);
+	while (len > 0) {
+		aktlen = min(SDD_BSIZE, len);
 		trys = 0;
 		do {
-			if (trys && !(trys & 15)){
+			if (trys && !(trys & 15)) {
 				if (debug) {
-					(void)putc('>', stderr);
-					(void)fflush(stderr);
+					(void) putc('>', stderr);
+					(void) fflush(stderr);
 				}
-				(void) riseek((off_t)(ifsize() - BSIZE));
-				(void) rread(rdbuf, (int) aktlen);
-			}
-			else if (trys > 0 && (trys == 2 || ! (trys & 7))) {
+				(void) roseek((off_t)(ifsize() - SDD_BSIZE));
+/* XXX we would need to read the output file here - open with "r" ??? */
+/*				(void) rread(rdbuf, (int) aktlen);*/
+
+			} else if (trys > 0 && (trys == 2 || ! (trys & 7))) {
 				if (debug) {
-					(void)putc('<', stderr);
-					(void)fflush(stderr);
+					(void) putc('<', stderr);
+					(void) fflush(stderr);
 				}
-				(void) riseek((off_t)0);
-				(void) rread(rdbuf, (int) aktlen);
+				(void) roseek((off_t)0);
+/* XXX we would need to read the output file here - open with "r" ??? */
+/*				(void) rread(rdbuf, (int) aktlen);*/
 			}
 			if (debug) {
-				(void)putc(';', stderr);
-				(void)fflush(stderr);
+				(void) putc(';', stderr);
+				(void) fflush(stderr);
 			}
 			(void) roseek(pos);
 			cnt = rwrite(buf, (int) aktlen);
+			if (cnt < 0) {
+				err = geterrno();
+#ifdef	ECONNRESET
+				if (err == EPIPE || err == ECONNRESET)
+#else
+				if (err == EPIPE)
+#endif
+					break;
+				err = 0;
+			}
 		} while (cnt < 0 && trys++ < try);
-		if(cnt < 0) {
+		if (cnt < 0) {
 			if (progress || debug) {
-				(void)putc('\n', stderr);
-				(void)fflush(stderr);
+				(void) putc('\n', stderr);
+				(void) fflush(stderr);
 			}
 			errmsgno(EX_BAD, "Block %lld not written correctly.\n",
-								(Llong)pos/BSIZE);
+						(Llong)pos/(Llong)SDD_BSIZE);
+			if (err != 0)
+				term((int) err);
 			cnt = aktlen;
 			writeerrs++;
-		}
-		else if (cnt == 0)	/* EOF */
+
+		} else if (cnt == 0)	/* EOF */
 			break;
 		buf += cnt;
 		len -= cnt;
@@ -1075,14 +1153,14 @@ fill(bp, start, end)
 	register char *p = &bp[start];
 	register char *ep = &bp[end];
 
-	while(p < ep)
+	while (p < ep)
 		*p++ = '\0';
 #else
 	fillbytes(&bp[start], end-start, '\0');
 #endif
 }
 
-#define	DO8(a)  a;a;a;a;a;a;a;a;
+#define	DO8(a)  a; a; a; a; a; a; a; a;
 
 LOCAL void
 swabb(bp, cnt)
@@ -1112,7 +1190,7 @@ lcase(bp, cnt)
 	register char	*bp;
 	register long	cnt;
 {
-	while(--cnt >= 0) {
+	while (--cnt >= 0) {
 		if (*bp >= 'A' && *bp <= 'Z')
 			*bp += 'a' - 'A';
 		bp++;
@@ -1124,7 +1202,7 @@ ucase(bp, cnt)
 	register char	*bp;
 	register long	cnt;
 {
-	while(--cnt >= 0) {
+	while (--cnt >= 0) {
 		if (*bp >= 'a' && *bp <= 'z')
 			*bp -= 'a' - 'A';
 		bp++;
@@ -1139,7 +1217,7 @@ block(bp, cnt)
 	register long	ocnt;
 
 	ocnt = cnt;
-	while(--cnt >= 0) {
+	while (--cnt >= 0) {
 		bp++;
 	}
 	return (ocnt);
@@ -1153,7 +1231,7 @@ unblock(bp, cnt)
 	register long	ocnt;
 
 	ocnt = cnt;
-	while(--cnt >= 0) {
+	while (--cnt >= 0) {
 		bp++;
 	}
 	return (ocnt);
@@ -1187,6 +1265,29 @@ LOCAL void
 term(ret)
 	int	ret;
 {
+	if (rmtout) {
+#ifdef	USE_REMOTE
+		/*
+		 * Cannot happen in non remote versions.
+		 */
+		if (rmtclose(rmtofd) < 0)
+			ret = geterrno();
+#endif
+	} else {
+#ifdef	HAVE_FSYNC
+		int	cnt = 0;
+
+		do {
+			if (fsync(ofd) != 0)
+				ret = geterrno();
+
+			if (ret == EINVAL)
+				ret = 0;
+		} while (ret == EINTR && ++cnt < 10);
+#endif
+		if (close(ofd) != 0 && ret == 0)
+			ret = geterrno();
+	}
 	prstats();
 	exit(ret);
 }
@@ -1231,20 +1332,20 @@ prstats()
 	iper = ((ibytes&1023)<<10)/10485;
 	oper = ((obytes&1023)<<10)/10485;
 
-	if (progress || debug) (void)putc('\n', stderr);
+	if (progress || debug) (void) putc('\n', stderr);
 	if (readerrs)
 		errmsgno(EX_BAD, "%lld %s(s) not read correctly.\n",
-					 readerrs, noseek?"Record":"Block");
+					readerrs, noseek?"Record":"Block");
 	if (writeerrs)
 		errmsgno(EX_BAD, "%lld %s(s) not written correctly.\n",
 					writeerrs, noseek?"Record":"Block");
 
 	errmsgno(EX_BAD,
-		"Read  %lld records + %lld bytes (total of %lld bytes = %lld.%02dk).\n",
-		irec, iparts, ibytes, ikbytes, iper);
+	"Read  %lld records + %lld bytes (total of %lld bytes = %lld.%02dk).\n",
+	irec, iparts, ibytes, ikbytes, iper);
 	errmsgno(EX_BAD,
-		"Wrote %lld records + %lld bytes (total of %lld bytes = %lld.%02dk).\n",
-		orec, oparts, obytes, okbytes, oper);
+	"Wrote %lld records + %lld bytes (total of %lld bytes = %lld.%02dk).\n",
+	orec, oparts, obytes, okbytes, oper);
 
 	if (flags & NULLIN) {
 		irec = savirec;
@@ -1342,16 +1443,19 @@ getopts(ac, av)
 	if (help)
 		usage(0);
 	if (prvers) {
-		printf("sdd %s (%s-%s-%s)\n\n", "1.31", HOST_CPU, HOST_VENDOR, HOST_OS);
-		printf("Copyright (C) 1984-2001 Jrg Schilling\n");
-		printf("This is free software; see the source for copying conditions.  There is NO\n");
-		printf("warranty; not even for MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.\n");
+		printf("sdd %s (%s-%s-%s)\n\n", "1.52",
+					HOST_CPU, HOST_VENDOR, HOST_OS);
+		printf("Copyright (C) 1984-2004 Jrg Schilling\n");
+		printf("This is free software; see the source for copying ");
+		printf("conditions.  There is NO\n");
+		printf("warranty; not even for MERCHANTABILITY or ");
+		printf("FITNESS FOR A PARTICULAR PURPOSE.\n");
 		exit(0);
 	}
 	cac = ac - 1;
 	cav = av + 1;
 	if (getfiles(&cac, &cav, opts) != 0) {
-		errmsgno(EX_BAD, "Bad Argument: '%s'\n",cav[0]);
+		errmsgno(EX_BAD, "Bad Argument: '%s'\n", cav[0]);
 		usage(EX_BAD);
 	}
 
@@ -1393,7 +1497,7 @@ getopts(ac, av)
 		usage(EX_BAD);
 	}
 	if (bs == 0)
-		bs = BSIZE;
+		bs = SDD_BSIZE;
 	if (ibs == 0)
 		ibs = bs;
 	if (obs == 0)
@@ -1408,7 +1512,7 @@ getopts(ac, av)
 		oseek = seek;
 	if (iskip || oskip) {
 		errmsgno(EX_BAD, "skip not implemented.\n");
-		usage (EX_BAD);
+		usage(EX_BAD);
 	}
 	if (fillflg)
 		flags |= FILL;
@@ -1424,7 +1528,7 @@ getopts(ac, av)
 	}
 	if (blkflg || ublkflg) {
 		errmsgno(EX_BAD, "block/unblock not implemented.\n");
-		usage (EX_BAD);
+		usage(EX_BAD);
 	}
 	if (lcflg && ucflg) {
 		errmsgno(EX_BAD, "Can't lcase and ucase.\n");
@@ -1472,33 +1576,39 @@ number(arg, retp)
 	if (*arg == '\0') {
 		*retp = -1;
 	} else if (*(arg = astoll(arg, &val))) {
-		if (*arg == 'p' || *arg == 'P') {
+		if (*arg == 'e' || *arg == 'E') {
+			val *= (1024*1024);
+			val *= (1024*1024);
+			val *= (1024*1024);
+			arg++;
+
+		} else if (*arg == 'p' || *arg == 'P') {
 			val *= (1024*1024);
 			val *= (1024*1024*1024);
 			arg++;
-		}
-		if (*arg == 't' || *arg == 'T') {
+
+		} else if (*arg == 't' || *arg == 'T') {
 			val *= (1024*1024);
 			val *= (1024*1024);
 			arg++;
-		}
-		if (*arg == 'g' || *arg == 'G') {
+
+		} else if (*arg == 'g' || *arg == 'G') {
 			val *= (1024*1024*1024);
 			arg++;
-		}
-		if (*arg == 'm' || *arg == 'M') {
+
+		} else if (*arg == 'm' || *arg == 'M') {
 			val *= (1024*1024);
 			arg++;
-		}
-		else if (*arg == 'k' || *arg == 'K') {
+
+		} else if (*arg == 'k' || *arg == 'K') {
 			val *= 1024;
 			arg++;
-		}
-		else if (*arg == 'b' || *arg == 'B') {
+
+		} else if (*arg == 'b' || *arg == 'B') {
 			val *= 512;
 			arg++;
-		}
-		else if (*arg == 'w' || *arg == 'W') {
+
+		} else if (*arg == 'w' || *arg == 'W') {
 			val *= 2;
 			arg++;
 		}
@@ -1521,7 +1631,9 @@ getnum(arg, valp)
 	llval = number(arg, &ret);
 	*valp = llval;
 	if (*valp != llval) {
-		errmsgno(EX_BAD, "Value %lld is too large for data type 'long'.\n", llval);
+		errmsgno(EX_BAD,
+			"Value %lld is too large for data type 'long'.\n",
+									llval);
 		ret = -1;
 	}
 	return (ret);
@@ -1546,7 +1658,7 @@ usage(ex)
 Usage:	sdd [option=value] [-flag]\n\
 Options:\n\
 ");
-	error ("\
+	error("\
 	if=name		  Read  input from name instead of stdin\n\
 	of=name		  Write output to name instead of stdout\n\
 	-inull		  Do not read input from file (use null char's)\n\
@@ -1587,13 +1699,13 @@ openremote(filename, iosize)
 
 #ifdef	USE_REMOTE
 	if ((remfn = rmtfilename(filename)) != NULL) {
-		rmthostname(host, filename, sizeof(host));
+		rmthostname(host, sizeof (host), filename);
 
 		if (debug)
 			errmsgno(EX_BAD, "Remote: %s Host: %s file: %s\n",
 							filename, host, remfn);
 
-		if ((remfd = rmtgetconn(host, iosize)) < 0)
+		if ((remfd = rmtgetconn(host, iosize, 0)) < 0)
 			comerrno(EX_BAD, "Cannot get connection to '%s'.\n",
 				/* errno not valid !! */		host);
 	}
@@ -1622,7 +1734,7 @@ ropenfile(rfd, name, mode)
 #endif
 }
 
-LOCAL int 
+LOCAL int
 rread(buf, cnt)
 	void	*buf;
 	int	cnt;
@@ -1634,7 +1746,7 @@ rread(buf, cnt)
 	return (_niread(ifd, buf, cnt));
 }
 
-LOCAL int 
+LOCAL int
 rwrite(buf, cnt)
 	void	*buf;
 	int	cnt;
